# Copyright (c) 2019 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Sequence, Union, Optional, Dict, Any

from attr import dataclass
import attr

from .formatted_string import FormattedString, EntityType


@dataclass
class Entity:
    type: EntityType
    offset: int
    length: int
    extra_info: Dict[str, Any] = attr.ib(factory=dict)

    def copy(self) -> 'Entity':
        return attr.evolve(self)

    def adjust_offset(self, offset: int, max_length: int = -1) -> Optional['Entity']:
        entity = attr.evolve(self, offset=self.offset + offset)
        if entity.offset < 0:
            entity.offset = 0
        elif entity.offset > max_length > -1:
            return None
        elif entity.offset + entity.length > max_length > -1:
            entity.length = max_length - entity.offset
        return entity


class EntityString(FormattedString):
    text: str
    entities: List[Entity]

    def __init__(self, text: str = "", entities: List[Entity] = None) -> None:
        self.text = text
        self.entities = entities or []

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text='{self.text}', entities={self.entities})"

    def __str__(self) -> str:
        return self.text

    def _offset_entities(self, offset: int) -> 'EntityString':
        if offset == 0:
            return self
        self.entities = [entity.adjust_offset(offset, len(self.text))
                         for entity in self.entities if entity
                         if entity.offset + offset <= len(self.text)]
        return self

    def append(self, *args: Union[str, 'FormattedString']) -> 'EntityString':
        for msg in args:
            if isinstance(msg, EntityString):
                self.entities += [entity.adjust_offset(len(self.text)) for entity in msg.entities]
                self.text += msg.text
            else:
                self.text += str(msg)
        return self

    def prepend(self, *args: Union[str, 'FormattedString']) -> 'EntityString':
        for msg in args:
            if isinstance(msg, EntityString):
                self.text = msg.text + self.text
                self.entities = msg.entities + [entity.adjust_offset(len(msg.text))
                                                for entity in self.entities]
            else:
                text = str(msg)
                self.text = text + self.text
                self.entities = [entity.adjust_offset(len(text)) for entity in self.entities]
        return self

    def format(self, entity_type: EntityType, offset: int = None, length: int = None, **kwargs
               ) -> 'EntityString':
        self.entities.append(Entity(type=entity_type, offset=offset or 0,
                                    length=length or len(self.text), extra_info=kwargs))
        return self

    def trim(self) -> 'EntityString':
        orig_len = len(self.text)
        self.text = self.text.lstrip()
        diff = orig_len - len(self.text)
        self.text = self.text.rstrip()
        self._offset_entities(-diff)
        return self

    def split(self, separator, max_items: int = -1) -> List['EntityString']:
        text_parts = self.text.split(separator, max_items - 1)
        output: List[EntityString] = []

        offset = 0
        for part in text_parts:
            msg = type(self)(part)
            for entity in self.entities:
                start_in_range = len(part) > entity.offset - offset >= 0
                end_in_range = len(part) >= entity.offset - offset + entity.length > 0
                if start_in_range and end_in_range:
                    msg.entities.append(entity.adjust_offset(-offset))
            output.append(msg)

            offset += len(part)
            offset += len(separator)

        return output

    @classmethod
    def join(cls, items: Sequence[Union[str, 'EntityString']],
             separator: str = " ") -> 'EntityString':
        main = cls()
        for msg in items:
            if not isinstance(msg, EntityString):
                msg = cls(text=str(msg))
            main.entities += [entity.adjust_offset(len(main.text)) for entity in msg.entities]
            main.text += msg.text + separator
        if len(separator) > 0:
            main.text = main.text[:-len(separator)]
        return main



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge
# Copyright (C) 2019 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Iterable
from http.cookies import SimpleCookie

from sqlalchemy import Column, String, PickleType
from sqlalchemy.engine.result import RowProxy

from mautrix.types import UserID
from mautrix.bridge.db.base import Base


class User(Base):
    __tablename__ = "user"

    mxid: UserID = Column(String(255), primary_key=True)
    session: SimpleCookie = Column(PickleType, nullable=True)
    fbid: str = Column(String(255), nullable=True)

    @classmethod
    def scan(cls, row: RowProxy) -> 'User':
        mxid, session, fbid = row
        return cls(mxid=mxid, session=session, fbid=fbid)

    @classmethod
    def all(cls) -> Iterable['User']:
        return cls._select_all()

    @classmethod
    def get_by_fbid(cls, fbid: str) -> Optional['User']:
        return cls._select_one_or_none(cls.c.fbid == fbid)

    @classmethod
    def get_by_mxid(cls, mxid: UserID) -> Optional['User']:
        return cls._select_one_or_none(cls.c.mxid == mxid)

    @property
    def _edit_identity(self):
        return self.c.mxid == self.mxid

    def insert(self) -> None:
        with self.db.begin() as conn:
            conn.execute(self.t.insert().values(mxid=self.mxid, session=self.session,
                                                fbid=self.fbid))



import setuptools
import glob
import mautrix_facebook

try:
    long_desc = open("README.md").read()
except IOError:
    long_desc = "Failed to read README.md"

setuptools.setup(
    name="mautrix-facebook",
    version=mautrix_facebook.__version__,
    url="https://github.com/tulir/mautrix-facebook",

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="A Matrix-Facebook Messenger puppeting bridge.",
    long_description=long_desc,
    long_description_content_type="text/markdown",

    packages=setuptools.find_packages(),

    install_requires=[
        "aiohttp>=3.0.1,<4",
        "mautrix>=0.4.0.dev51,<0.5.0",
        "ruamel.yaml>=0.15.94,<0.16",
        "commonmark>=0.8,<0.9",
        "python-magic>=0.4,<0.5",
        "fbchat-asyncio>=0.2.0,<0.3.0",
        "SQLAlchemy>=1.2,<2",
        "alembic>=1,<2",
    ],

    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points="""
        [console_scripts]
        mautrix-facebook=mautrix_facebook.__main__:main
    """,
    data_files=[
        (".", ["example-config.yaml", "alembic.ini"]),
        ("alembic", ["alembic/env.py"]),
        ("alembic/versions", glob.glob("alembic/versions/*.py"))
    ],
)



from mautrix.bridge.commands import SECTION_GENERAL
from .handler import (CommandProcessor, CommandHandler, CommandEvent, command_handler,
                      SECTION_AUTH)
from .auth import login, enter_2fa_code
